@@ -26,6 +26,8 @@ class Socket(object):
             self.queue.task_done()
         except self.server.async['queue'].Empty:
             raise IOError()
+        if packets == [None]:
+            return []
         try:
             packets.append(self.queue.get(block=False))
             self.queue.task_done()
@@ -35,14 +37,16 @@ class Socket(object):
 
     def receive(self, pkt):
         """Receive packet from the client."""
-        self.server.logger.info('%s: Received packet %s with %s', self.sid,
-                                packet.packet_names[pkt.packet_type],
-                                pkt.data)
+        self.server.logger.info('%s: Received packet %s data %s',
+                                self.sid, packet.packet_names[pkt.packet_type],
+                                pkt.data if not isinstance(pkt.data, bytes)
+                                else '<binary>')
         if pkt.packet_type == packet.PING:
             self.last_ping = time.time()
             self.send(packet.Packet(packet.PONG, pkt.data))
         elif pkt.packet_type == packet.MESSAGE:
-            self.server._trigger_event('message', self.sid, pkt.data)
+            self.server._trigger_event('message', self.sid, pkt.data,
+                                       async=self.server.async_handlers)
         elif pkt.packet_type == packet.UPGRADE:
             self.send(packet.Packet(packet.NOOP))
         else:
@@ -52,15 +56,16 @@ class Socket(object):
         """Send a packet to the client."""
         if self.closed:
             raise IOError('Socket is closed')
-        if time.time() - self.last_ping > self.server.ping_interval * 5 / 4:
+        if time.time() - self.last_ping > self.server.ping_timeout:
             self.server.logger.info('%s: Client is gone, closing socket',
                                     self.sid)
             self.close(wait=False, abort=True)
             return
         self.queue.put(pkt)
-        self.server.logger.info('%s: Sending packet %s with %s', self.sid,
-                                packet.packet_names[pkt.packet_type],
-                                pkt.data)
+        self.server.logger.info('%s: Sending packet %s data %s',
+                                self.sid, packet.packet_names[pkt.packet_type],
+                                pkt.data if not isinstance(pkt.data, bytes)
+                                else '<binary>')
 
     def handle_get_request(self, environ, start_response):
         """Handle a long-polling GET request from the client."""
@@ -93,7 +98,7 @@ class Socket(object):
 
     def close(self, wait=True, abort=False):
         """Close the socket connection."""
-        self.server._trigger_event('disconnect', self.sid)
+        self.server._trigger_event('disconnect', self.sid, async=False)
         if not abort:
             self.send(packet.Packet(packet.CLOSE))
         self.closed = True
@@ -104,6 +109,10 @@ class Socket(object):
         """Upgrade the connection from polling to websocket."""
         if self.upgraded:
             raise IOError('Socket has been upgraded already')
+        if self.server.async['websocket'] is None or \
+                self.server.async['websocket_class'] is None:
+            # the selected async mode does not support websocket
+            return self.server._bad_request()
         websocket_class = getattr(self.server.async['websocket'],
                                   self.server.async['websocket_class'])
         ws = websocket_class(self._websocket_handler)
@@ -121,37 +130,42 @@ class Socket(object):
                                         always_bytes=False):
                 self.server.logger.info(
                     '%s: Failed websocket upgrade, no PING packet', self.sid)
-                return
+                return []
             ws.send(packet.Packet(
                 packet.PONG,
                 data=six.text_type('probe')).encode(always_bytes=False))
             self.send(packet.Packet(packet.NOOP))
 
             pkt = ws.wait()
-            if pkt != packet.Packet(packet.UPGRADE).encode(always_bytes=False):
+            decoded_pkt = packet.Packet(encoded_packet=pkt)
+            if decoded_pkt.packet_type != packet.UPGRADE:
                 self.upgraded = False
                 self.server.logger.info(
-                    '%s: Failed websocket upgrade, no UPGRADE packet',
-                    self.sid)
-                return
+                    ('%s: Failed websocket upgrade, expected UPGRADE packet, '
+                     'received %s instead.'),
+                    self.sid, pkt)
+                return []
             self.upgraded = True
         else:
             self.connected = True
             self.upgraded = True
 
+        # start separate writer thread
         def writer():
             while True:
                 try:
                     packets = self.poll()
                 except IOError:
                     break
+                if not packets:
+                    # empty packet list returned -> connection closed
+                    break
                 try:
                     for pkt in packets:
                         ws.send(pkt.encode(always_bytes=False))
                 except:
                     break
-
-        writer_task = self.server.start_background_task(writer)
+        self.server.start_background_task(writer)
 
         self.server.logger.info(
             '%s: Upgrade to websocket successful', self.sid)
@@ -162,6 +176,7 @@ class Socket(object):
             except:
                 break
             if p is None:
+                # connection closed by client
                 break
             if isinstance(p, six.text_type):  # pragma: no cover
                 p = p.encode('utf-8')
@@ -170,5 +185,8 @@ class Socket(object):
                 self.receive(pkt)
             except ValueError:
                 pass
-        self.close(wait=False, abort=True)
-        writer_task.join()
+
+        self.close(wait=True, abort=True)
+        self.queue.put(None)  # unlock the writer task so that it can exit
+
+        return []
